from .segment import Segment


class Envelope(object):
    def __init__(self):
        self.header = Segment()
        self.trailer = Segment()
        self.body = []

    def format_as_edi(self, document_configuration):
        """
        Format the envelope as an EDI string.
        :param document_configuration: configuration for formatting.
        :return: document as a string of EDI.
        """
        document = self.header.format_as_edi(document_configuration)
        document += self.__format_body_as_edi(document_configuration)
        document += self.trailer.format_as_edi(document_configuration)
        return document

    def __format_body_as_edi(self, document_configuration):
        """
        Format the body of the envelope as an EDI string.
        This calls format_as_edi in all the children.
        :param document_configuration: configuration for formatting.
        :return: document as a string of EDI.
        """
        document = ""
        for item in self.body:
            document += item.format_as_edi(document_configuration)
        return document

    def validate(self, report):
        """
        Performs validation of the envelope and its components.
        :param report: the validation report to append errors.
        """
        self.header.validate(report)
        self.__validate_body(report)
        self.trailer.validate(report)

    def __validate_body(self, report):
        """
        Validates each of the children of the envelope.
        :param report: the validation report to append errors.
        """
        for item in self.body:
            item.validate(report)

    def number_of_segments(self):
        return len(self.body)


class InterchangeEnvelope(Envelope):
    def __init__(self):
        Envelope.__init__(self)
        self.groups = self.body


class GroupEnvelope(Envelope):
    def __init__(self):
        Envelope.__init__(self)
        self.transaction_sets = self.body


class TransactionSetEnvelope(Envelope):
    def __init__(self):
        Envelope.__init__(self)
        self.transaction_body = self.body

    def number_of_segments(self):
        header_trailer_count = 2
        return len(self.transaction_body) + header_trailer_count



from badX12.settings import DocumentSettings
from .element import Element


class Segment(object):
    def __init__(self):
        self.fieldCount = 0
        self.fields = []
        self.id = Element()
        self.element_separator = DocumentSettings.element_separator
        self.segment_terminator = DocumentSettings.segment_terminator
        self.sub_element_separator = DocumentSettings.sub_element_separator

    def validate(self, report):
        """
        Validate the segment by validating all elements.
        :param report: the validation report to append errors.
        """
        for field in self.fields:
            field.validate(report)

    def format_as_edi(self, document_configuration):
        """Format the segment into an EDI string"""
        self.element_separator = document_configuration.element_separator
        self.segment_terminator = document_configuration.segment_terminator
        self.sub_element_separator = document_configuration.sub_element_separator
        return str(self)

    def __str__(self):
        """Return the segment as a string"""
        out = ""
        if self.__all_fields_empty():
            return out
        out = self.__get_fields_as_string(out)
        return out

    def __all_fields_empty(self):
        """determine if all fields are empty"""
        for field in self.fields:
            if field.content != "":
                return False
        return True

    def __get_fields_as_string(self, out):
        """processes all the fields in the segment and returns the string representation"""
        for index, field in enumerate(self.fields):
            if field.content:
                out += str(field)
                out = self.__add_separators_to_fields(index, out)
        out = self.__add_segment_terminator(out)

        return out

    def __add_separators_to_fields(self, index, out):
        """adds field separators to all the element strings"""
        if index < self.fieldCount:
            # if the next field is required add the separator
            if self.fields[index + 1].required:
                out += self.element_separator
            # if the next field is optional but has content add the separator
            elif self.fields[index + 1].content:
                out += self.element_separator
            # finally if the next field is not the last element add the separator
            elif index + 1 != self.fieldCount:
                out += self.element_separator
        return out

    def __add_segment_terminator(self, out):
        """Adds the segment terminator to the segment string"""
        out += self.segment_terminator
        return out

    def __repr__(self):
        return f"""
            Segment(
                Field Count => {self.fieldCount}
                Fields => {self.fields}
                Element Separator => {self.element_separator}
                Segment Terminator => {self.segment_terminator}
                Sub Element Separator => {self.sub_element_separator}
            )
        """


name = 'badX12'
version = '0.1.0'
author = 'Alberto J. Marin'
author_email = 'alberto@ajmar.in'
homepage = 'https://github.com/git-albertomarin/badX12'
default_user_agent = f'{name}/{version} (+{homepage})'



from badX12.common import TransactionSetEnvelope, Element, Segment
from badX12.common.errors import IDMismatchError, SegmentCountError


class TransactionSet(TransactionSetEnvelope):
    """An EDI X12 transaction set"""

    def __init__(self):
        TransactionSetEnvelope.__init__(self)
        self.header = TransactionSetHeader()
        self.trailer = TransactionSetTrailer()

    def validate(self, report):
        """
        Validate the envelope
        :param report: the validation report to append errors.
        """
        super(TransactionSetEnvelope, self).validate(report)
        self.__validate_control_ids(report)
        self.__validate_group_count(report)

    def __validate_control_ids(self, report):
        """
        Validate the control id match in the header and trailer
        :param report: the validation report to append errors.
        """
        if self.header.st02.content != self.trailer.se02.content:
            st02_desc = self.header.st02.description
            st02_name = self.header.st02.name
            se02_desc = self.trailer.se02.description
            se02_name = self.trailer.se02.name
            report.add_error(IDMismatchError(
                msg="The " + st02_desc + " in " + st02_name + " does not match " + se02_desc + " in " + se02_name,
                segment=self.header.id))

    def __validate_group_count(self, report):
        """
        Validate the actual group count matches the specified count.
        :param report: the validation report to append errors.
        """
        if int(self.trailer.se01.content) != self.number_of_segments():
            report.add_error(SegmentCountError(
                msg="The " + self.trailer.se01.description + " in " + self.trailer.se01.name +
                    " value of " + self.trailer.se01.content + " does not match the parsed count of " +
                    str(len(self.transaction_body)),
                segment=self.trailer.id))

    def __repr__(self):
        return f"""
            TransactionSet(
                Header => {self.header}
                Trailer => {self.trailer}
                Body => {self.body}
            )
        """


class TransactionSetHeader(Segment):
    """The transaction set header"""

    def __init__(self):
        Segment.__init__(self)
        self.fieldCount = 3

        self.id = Element(name="ST",
                          description="Transaction Set Header",
                          required=True, min_length=2, max_length=2, content="ST")
        self.fields.append(self.id)

        self.st01 = Element(name="ST01",
                            description="Transaction Set ID Code",
                            required=True, min_length=3, max_length=3, content="")
        self.fields.append(self.st01)

        self.st02 = Element(name="ST02",
                            description="Transaction Set Control Number",
                            required=True, min_length=4, max_length=9, content="")
        self.fields.append(self.st02)

        self.st03 = Element(name="ST03",
                            description="Implementation Convention Reference",
                            required=False, min_length=1, max_length=35, content="")
        self.fields.append(self.st03)

    def __repr__(self):
        return f"""
            TransactionSetHeader (
                Field Count => {self.fieldCount}
                Fields => {self.fields}
            )
        """


class TransactionSetTrailer(Segment):
    """The transaction set header"""

    def __init__(self):
        Segment.__init__(self)
        self.fieldCount = 2

        self.id = Element(name="SE",
                          description="Transaction Set Trailer",
                          required=True, min_length=2, max_length=2, content="SE")
        self.fields.append(self.id)

        self.se01 = Element(name="SE01",
                            description="Number of Included Segments",
                            required=True, min_length=1, max_length=6, content="")
        self.fields.append(self.se01)

        self.se02 = Element(name="SE02",
                            description="Transaction Set Control Number",
                            required=True, min_length=4, max_length=9, content="")
        self.fields.append(self.se02)



from badX12.common import Interchange
from badX12.settings import DocumentSettings


class DocumentConfiguration:

    def __init__(self, version, element_separator, segment_terminator, sub_element_separator):
        """Creates a new Edi Document Configuration"""
        self.version = version
        self.element_separator = element_separator
        self.segment_terminator = segment_terminator
        self.sub_element_separator = sub_element_separator


class ValidationReport:
    def __init__(self):
        self.error_list = []

    def add_error(self, error):
        self.error_list.append(error)

    def is_document_valid(self):
        return len(self.error_list) == 0


class EDIDocument:
    """
    An EDI X12 Document
    """

    def __init__(self):
        self.text = ""
        self.configuration = DocumentConfiguration(
            version=DocumentSettings.version,
            element_separator=DocumentSettings.element_separator,
            segment_terminator=DocumentSettings.segment_terminator,
            sub_element_separator=DocumentSettings.sub_element_separator
        )

        self.interchange = Interchange()

    def format_as_edi(self):
        """Format this document as EDI and return it as a string"""
        return self.interchange.format_as_edi(self.document_configuration)

    def validate(self):
        """Validate this document and return a validation report"""
        report = ValidationReport()
        self.interchange.validate(report)
        return report

    def __repr__(self):
        newline = '\n'

        return f"""
        EDIDocument(
            Text => {self.text}
            Document Text => ('{self.document_text}'...)
            Configuration => 
                {self.configuration}
            Interchange => 
                {self.interchange}
            Version => {self.version}
        )    
        """


"""
    badX12
    ~~~~~~~~~~~~

    A Python API for parsing ANSI ASC X12 files.
"""
from .__project__ import version, homepage, author
from .parsers import Parser

__all__ = ['Parser']

__version__ = version
__homepage__ = homepage
__author__ = author
__license__ = 'Apache-2.0'



from badX12.common import InterchangeHeader, Element, Segment
from badX12.common.errors import InvalidFileTypeError, SegmentTerminatorNotFoundError
from badX12.documents import EDIDocument
from ._group import Group, GroupHeader, GroupTrailer
from ._transaction_set import TransactionSet, TransactionSetHeader, TransactionSetTrailer
import os

class Parser:
    def __init__(self):
        """Create a new Parser"""
        self.edi_document = EDIDocument()
        self.document_text = ""

    def parse_document(self, document_text):
        """Parse the text document into an object
        :param document_text:  the text to parse into an EDI document.
        """
        if os.path.isfile(document_text):
            with open(document_text, 'r') as x12File:
                document_text = x12File.read().strip()

        self.document_text = document_text.replace('\n', '').strip()
        # attach the original document text to the document.
        self.edi_document.document_text = self.document_text

        if self.document_text.startswith(EDIDocument().interchange.header.id.name):
            self.__parse_interchange_header()
            self.__separate_and_route_segments()

        else:
            found_segment = self.document_text[:3]
            raise InvalidFileTypeError(segment=found_segment,
                                       msg="Expected Element Envelope: " + EDIDocument().interchange.header.id.name +
                                           " but found Element Envelope: " + found_segment +
                                           ".\n The length of the expected segment is: " + str(
                                           len(EDIDocument().interchange.header.id.name)) +
                                           " the length of the segment found was: " + str(len(found_segment)))

        return self.edi_document

    def __parse_interchange_header(self):
        """Parse the interchange header segment"""
        header = self.edi_document.interchange.header
        # The edi separator is always at position 4
        self.edi_document.configuration.element_separator = self.document_text[3:4]
        header_field_list = self.document_text.split(self.edi_document.configuration.element_separator)
        for index, isa in enumerate(header_field_list):
            if index == 12:
                self.edi_document.version = isa
            if index <= 15:
                header.fields[index].content = isa
            if index == 16:
                last_header_field = header_field_list[16]
                # the sub-element separator is always the first character in this element.
                header.isa16.content = last_header_field[0:1]
                if last_header_field[1:2]:
                    self.edi_document.configuration.segment_terminator = last_header_field[1:2]
                else:
                    raise SegmentTerminatorNotFoundError(
                        msg="The segment terminator is not present in the Interchange Header, can't parse file.")

    def __separate_and_route_segments(self):
        """Handles separating all the segments"""
        self.segment_list = self.document_text.split(self.edi_document.configuration.segment_terminator)
        for segment in self.segment_list:
            self.__route_segment_to_parser(segment)

    def __route_segment_to_parser(self, segment):
        """Take a generic segment and determine what segment to parse it as
        :param segment:
        """
        if segment.startswith(InterchangeHeader().id.name):
            pass
        elif segment.startswith(GroupHeader().id.name):
            self.__parse_group_header(segment)
        elif segment.startswith(GroupTrailer().id.name):
            self.__parse_group_trailer(segment)
        elif segment.startswith(TransactionSetHeader().id.name):
            self.__parse_transaction_set_header(segment)
        elif segment.startswith(TransactionSetTrailer().id.name):
            self.__parse_transaction_set_trailer(segment)
        elif segment.startswith(EDIDocument().interchange.trailer.id.name):
            self.__parse_interchange_trailer(segment)
        else:
            self.__parse_unknown_body(segment)

    def __parse_segment(self, segment, segment_field_list):
        """Generically parse segments
        :param segment: the segment to insert the values.
        :param segment_field_list: the list of segments to parse.
        """
        for index, value in enumerate(segment_field_list):
            segment.fields[index].content = value

    def __create_generic_element(self, index, value):
        """
        Create a generic element based on the data found. Populate all the
        fields so that validation will pass.
        :param index: the position of the element for providing a name.
        :param value: the content for the element being created.
        :return: a generic element.
        """
        element = Element()
        element.name = "GEN" + str(index)
        element.content = value
        element.description = "A generic element created by the parser"
        element.required = False
        length = len(value)
        element.minLength = length
        element.maxLength = length
        return element

    def __parse__unknown_segment(self, segment, segmentFieldList):
        """Generically parse unknown segments by creating a
        new element and appending it to the segment.
        :param segment: the segment to append the values.
        :param segmentFieldList: the list of segments to parse.
        """
        for index, value in enumerate(segmentFieldList):
            element = self.__create_generic_element(index, value)
            segment.fields.append(element)

    def __parse_group_header(self, segment):
        """Parse the group header"""
        self.current_group = Group()
        header = GroupHeader()
        header_field_list = segment.split(self.edi_document.configuration.element_separator)
        self.__parse_segment(header, header_field_list)
        self.current_group.header = header

    def __parse_group_trailer(self, segment):
        """Parse the group trailer"""
        trailer = GroupTrailer()
        trailer_field_list = segment.split(self.edi_document.configuration.element_separator)
        self.__parse_segment(trailer, trailer_field_list)
        self.current_group.trailer = trailer
        self.edi_document.interchange.groups.append(self.current_group)

    def __parse_interchange_trailer(self, segment):
        """Parse the interchange trailer segment"""
        trailer = self.edi_document.interchange.trailer
        trailer_field_list = segment.split(self.edi_document.configuration.element_separator)
        self.__parse_segment(trailer, trailer_field_list)

    def __parse_transaction_set_header(self, segment):
        """Parse transaction set header
        Creates a new transaction set and set it as the current transaction set.
        """
        self.current_transaction = TransactionSet()
        transaction_header = TransactionSetHeader()
        header_field_list = segment.split(self.edi_document.configuration.element_separator)
        self.__parse_segment(transaction_header, header_field_list)
        self.current_transaction.header = transaction_header

    def __parse_transaction_set_trailer(self, segment):
        """Parse the transaction set trailer.
        Adds the completed transaction to a edi document.
        """
        transaction_trailer = TransactionSetTrailer()
        trailer_field_list = segment.split(self.edi_document.configuration.element_separator)
        self.__parse_segment(transaction_trailer, trailer_field_list)
        self.current_transaction.trailer = transaction_trailer
        self.current_group.transaction_sets.append(self.current_transaction)

    def __parse_unknown_body(self, segment):
        if segment:
            generic_segment = Segment()
            generic_field_list = segment.split(self.edi_document.configuration.element_separator)
            self.__parse__unknown_segment(generic_segment, generic_field_list)
            try:
                self.current_transaction.transaction_body.append(generic_segment)
            except AttributeError:
                pass



from badX12.common import GroupEnvelope, Element, Segment
from badX12.common.errors import SegmentCountError, IDMismatchError


class Group(GroupEnvelope):
    """An EDI X12 groups"""

    def __init__(self):
        GroupEnvelope.__init__(self)
        self.header = GroupHeader()
        self.trailer = GroupTrailer()

    def validate(self, report):
        """
        Validate the group envelope
        :param report: the validation report to append errors.
        """
        super(GroupEnvelope, self).validate(report)
        self.__validate_control_ids(report)
        self.__validate_group_count(report)

    def __validate_control_ids(self, report):
        """
        Validate the control id match in the header and trailer
        :param report: the validation report to append errors.
        """
        if self.header.gs06.content != self.trailer.ge02.content:
            gs06_desc = self.header.gs06.description
            gs06_name = self.header.gs06.name
            ge02_desc = self.trailer.ge02.description
            ge02_name = self.trailer.ge02.name
            report.add_error(IDMismatchError(
                msg="The " + gs06_desc + " in " + gs06_name + " does not match " + ge02_desc + " in " + ge02_name,
                segment=self.header.id))

    def __validate_group_count(self, report):
        """
        Validate the actual group count matches the specified count.
        :param report: the validation report to append errors.
        """
        if int(self.trailer.ge01.content) != len(self.transaction_sets):
            report.add_error(SegmentCountError(
                msg="The " + self.trailer.ge01.description + " in " + self.trailer.ge01.name +
                    " value of " + self.trailer.ge01.content + " does not match the parsed count of " +
                    str(len(self.transaction_sets)),
                segment=self.trailer.id))

    def __repr__(self):
        return f"""
            Group(
                Header => {self.header}
                Trailer => {self.trailer}
                Body => {self.body}
                Transaction Sets => {self.transaction_sets}
            )
        """


class GroupHeader(Segment):
    """An EDI X12 groups header"""

    def __init__(self):
        Segment.__init__(self)
        self.fieldCount = 8

        self.id = Element(name="GS",
                          description="Functional Group Header Code",
                          required=True, min_length=2, max_length=2, content="GS")
        self.fields.append(self.id)

        self.gs01 = Element(name="GS01",
                            description="Functional Identifier Code",
                            required=True, min_length=2, max_length=2, content="")
        self.fields.append(self.gs01)

        self.gs02 = Element(name="GS02",
                            description="Application Senders Code",
                            required=True, min_length=2, max_length=15, content="")
        self.fields.append(self.gs02)

        self.gs03 = Element(name="GS03",
                            description="Application Receiver Code",
                            required=True, min_length=2, max_length=15, content="")
        self.fields.append(self.gs03)

        self.gs04 = Element(name="GS04",
                            description="Group Date",
                            required=True, min_length=8, max_length=8, content="")
        self.fields.append(self.gs04)

        self.gs05 = Element(name="GS05",
                            description="Group Time",
                            required=True, min_length=4, max_length=4, content="")
        self.fields.append(self.gs05)

        self.gs06 = Element(name="GS06",
                            description="Group Control Number",
                            required=True, min_length=1, max_length=9, content="")
        self.fields.append(self.gs06)

        self.gs07 = Element(name="GS07",
                            description="Responsible Agency Code",
                            required=True, min_length=1, max_length=2, content="")
        self.fields.append(self.gs07)

        self.gs08 = Element(name="GS08",
                            description="Version Indicator ID Code",
                            required=True, min_length=1, max_length=12, content="")
        self.fields.append(self.gs08)

    def __repr__(self):
        return f"""
            GroupHeader (
                Field Count => {self.fieldCount},
                Fields => {[val.content for val in self.fields]}
                Id => {self.id.name}
                Element Separator => {self.element_separator}
                Segment Terminator => {self.segment_terminator}
                Sub Element Separator => {self.sub_element_separator}
                GS01 => {self.gs01.content}
                GS02 => {self.gs02.content}
                GS03 => {self.gs03.content}
                GS04 => {self.gs04.content}
                GS05 => {self.gs05.content}
                GS06 => {self.gs06.content}
                GS07 => {self.gs07.content}
                GS08 => {self.gs08.content}
            )
        """


class GroupTrailer(Segment):
    """An EDI X12 groups trailer"""

    def __init__(self):
        Segment.__init__(self)
        self.fieldCount = 2

        self.id = Element(name="GE",
                          description="Functional Group Trailer Identifier",
                          required=True, min_length=2, max_length=2, content="GE")
        self.fields.append(self.id)

        self.ge01 = Element(name="GE01",
                            description="Number of Transaction Sets",
                            required=True, min_length=1, max_length=6, content="")
        self.fields.append(self.ge01)

        self.ge02 = Element(name="GE02",
                            description="Group Control Number",
                            required=True, min_length=1, max_length=9, content="")
        self.fields.append(self.ge02)



from .envelope import InterchangeEnvelope
from .segment import Segment
from .element import Element
from .errors import IDMismatchError, SegmentCountError


class Interchange(InterchangeEnvelope):
    """An EDI X12 interchange"""

    def __init__(self):
        InterchangeEnvelope.__init__(self)
        self.header = InterchangeHeader()
        self.trailer = InterchangeTrailer()

    def validate(self, report):
        """
        Validate the envelope
        :param report: the validation report to append errors.
        """
        super(InterchangeEnvelope, self).validate(report)
        self.__validate_control_ids(report)
        self.__validate_group_count(report)

    def __validate_control_ids(self, report):
        """
        Validate the control id match in the header and trailer
        :param report: the validation report to append errors.
        """
        if self.header.isa13.content != self.trailer.iea02.content:
            isa13_desc = self.header.isa13.description
            isa13_name = self.header.isa13.name
            iea02_desc = self.trailer.iea02.description
            iea02_name = self.trailer.iea02.name
            report.add_error(IDMismatchError(
                msg="The " + isa13_desc + " in " + isa13_name + " does not match " + iea02_desc + " in " + iea02_name,
                segment=self.header.id))

    def __validate_group_count(self, report):
        """
        Validate the actual group count matches the specified count.
        :param report: the validation report to append errors.
        """
        if int(self.trailer.iea01.content) != len(self.groups):
            report.add_error(SegmentCountError(
                msg="The " + self.trailer.iea01.description + " in " + self.trailer.iea01.name +
                    " value of " + self.trailer.iea01.content + " does not match the parsed count of " +
                    str(len(self.groups)),
                segment=self.trailer.id))

    def __repr__(self):
        return f"""
            Interchange(
                Header => {self.header}
                Trailer => {self.trailer}
                Body => {self.body}
                Groups => {self.groups}
            )
        """


class InterchangeHeader(Segment):
    """An EDI X12 interchange header"""

    def __init__(self):
        Segment.__init__(self)
        self.fieldCount = 16

        self.id = Element(name="ISA",
                          description="Interchange Control Header Code",
                          required=True, min_length=3, max_length=3, content="ISA")
        self.fields.append(self.id)

        self.isa01 = Element(name="ISA01",
                             description="Authorization Information Qualifier",
                             required=True, min_length=2, max_length=2, content="")
        self.fields.append(self.isa01)

        self.isa02 = Element(name="ISA02",
                             description="Authorization Information",
                             required=True, min_length=10, max_length=10, content="")
        self.fields.append(self.isa02)

        self.isa03 = Element(name="ISA03",
                             description="Security Information Qualifier",
                             required=True, min_length=2, max_length=2, content="")
        self.fields.append(self.isa03)

        self.isa04 = Element(name="ISA04",
                             description="Security Information",
                             required=True, min_length=10, max_length=10, content="")
        self.fields.append(self.isa04)

        self.isa05 = Element(name="ISA05",
                             description="Interchange ID Qualifier",
                             required=True, min_length=2, max_length=2, content="")
        self.fields.append(self.isa05)

        self.isa06 = Element(name="ISA06",
                             description="Interchange Sender ID",
                             required=True, min_length=15, max_length=15, content="")
        self.fields.append(self.isa06)

        self.isa07 = Element(name="ISA07",
                             description="Interchange ID Qualifier",
                             required=True, min_length=2, max_length=2, content="")
        self.fields.append(self.isa07)

        self.isa08 = Element(name="ISA08",
                             description="Interchange Receiver ID",
                             required=True, min_length=15, max_length=15, content="")
        self.fields.append(self.isa08)

        self.isa09 = Element(name="ISA09",
                             description="Interchange Date",
                             required=True, min_length=6, max_length=6, content="")
        self.fields.append(self.isa09)

        self.isa10 = Element(name="ISA10",
                             description="Interchange Time",
                             required=True, min_length=4, max_length=4, content="")
        self.fields.append(self.isa10)

        self.isa11 = Element(name="ISA11",
                             description="Interchange Control Standard ID",
                             required=True, min_length=1, max_length=1, content="")
        self.fields.append(self.isa11)

        self.isa12 = Element(name="ISA12",
                             description="Interchange Control Version Number",
                             required=True, min_length=5, max_length=5, content="")
        self.fields.append(self.isa12)

        self.isa13 = Element(name="ISA13",
                             description="Interchange Control Number",
                             required=True, min_length=9, max_length=9, content="")
        self.fields.append(self.isa13)

        self.isa14 = Element(name="ISA14",
                             description="Acknowledgement Requested Flag",
                             required=True, min_length=1, max_length=1, content="")
        self.fields.append(self.isa14)

        self.isa15 = Element(name="ISA15",
                             description="Test Indicator",
                             required=True, min_length=1, max_length=1, content="")
        self.fields.append(self.isa15)

        self.isa16 = Element(name="ISA16",
                             description="Sub-element Separator",
                             required=True, min_length=1, max_length=1, content="")
        self.fields.append(self.isa16)

    def __repr__(self):
        return f"""
                InterchangeHeader(
                    Field Count => {self.fieldCount}
                    Fields => {[val.content for val in self.fields]}
                    Element Separator => {self.element_separator}
                    Segment Terminator => {self.segment_terminator}
                    Sub Element Separator => {self.sub_element_separator}
                    ISA01 => {self.isa01}
                    ISA02 => {self.isa02}
                    ISA03 => {self.isa03}
                    ISA04 => {self.isa04}
                    ISA05 => {self.isa05}
                    ISA06 => {self.isa06}
                    ISA07 => {self.isa07}
                    ISA08 => {self.isa08}
                    ISA09 => {self.isa09}
                    ISA10 => {self.isa10}
                    ISA11 => {self.isa11}
                    ISA12 => {self.isa12}
                    ISA13 => {self.isa13}
                    ISA14 => {self.isa14}
                    ISA15 => {self.isa15}
                    ISA16 => {self.isa16}
                )
        """


class InterchangeTrailer(Segment):
    """An EDI X12 interchange Trailer"""

    def __init__(self):
        Segment.__init__(self)
        self.fieldCount = 2

        self.id = Element(name="IEA",
                          description="Interchange Control Trailer Code",
                          required=True, min_length=3, max_length=3, content="")
        self.fields.append(self.id)

        self.iea01 = Element(name="IEA01",
                             description="Number of Included Groups",
                             required=True, min_length=1, max_length=5, content="")
        self.fields.append(self.iea01)

        self.iea02 = Element(name="IEA02",
                             description="Interchange Control Number",
                             required=True, min_length=1, max_length=9, content="")
        self.fields.append(self.iea02)

    def __repr__(self):
        return f"""
                InterchangeTrailer(
                    Field Count => {self.fieldCount}
                    Fields => {[val.content for val in self.fields]}
                    Element Separator => {self.element_separator}
                    Segment Terminator => {self.segment_terminator}
                    Sub Element Separator => {self.sub_element_separator}
                    IEA01 => {self.iea01}
                    IEA02 => {self.iea02}
                )
        """


from .errors import FieldValidationError


class Element(object):
    """A generic segment"""

    def __init__(self, name="", description="", required="", min_length="", max_length="", content=""):
        self.name = name
        self.description = description
        self.required = required
        self.minLength = min_length
        self.maxLength = max_length
        self.content = content

    def validate(self, report):
        """Validate the element"""
        if self.required or self.content != "":
            content_length = len(self.content)
            self.__is_field_too_short(content_length, report)
            self.__is_field_too_long(content_length, report)

    def __str__(self):
        if self.required:
            return str(self.content)

        if self.content != "":
            return str(self.content)

        return ""

    def __is_field_too_short(self, content_length, report):
        """
        Determine if the field content is too short.
        :param content_length: current content length.
        :param report: the validation report to append errors.
        """
        if content_length < self.minLength:
            report.add_error(
                FieldValidationError(msg="Field " + self.name + " is too short. Found " + str(content_length)
                                         + " characters, expected " + str(self.minLength) + " characters.",
                                     segment=self))

    def __is_field_too_long(self, content_length, report):
        """
        Determine if the field content is too long.
        :param content_length: current content length.
        :param report: the validation report to append errors.
        """
        if content_length > self.maxLength:
            report.add_error(
                FieldValidationError(msg="Field " + self.name + " is too long. Found " + str(content_length)
                                         + " characters, expected " + str(self.maxLength) + " characters.",
                                     segment=self))

    def __repr__(self):
        return f"""
            Element(
                Name => {self.name}
                Description => {self.description}
                Required => {self.required}
                Min. Length => {self.minLength}
                Max. Length => {self.maxLength}
                Content => {self.content}
            )
        """


#!/usr/bin/env python
from setuptools import setup, find_packages

from badX12.__project__ import author, author_email, version, homepage


def fread(filename):
    with open(filename) as f:
        return f.read()


setup(
    name='badX12',
    version=version,

    author=author,
    author_email=author_email,
    url=homepage,
    packages=find_packages(exclude=('tests', 'docs')),
    description=(
        'A Python API for parsing ANSI ASC X12 files.'
    ),
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    long_description=fread('README.rst'),
    license=fread('LICENSE'),

    install_requires=[
    ],

    project_urls={
        'Bug Tracker': f'{homepage}/issues',
        'Documentation': f'{homepage}/tree/master/docs',
        'Source Code': f'{homepage}',
    },

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    keywords='x12 edi parser'
)



class DefaultSettings(object):
    """Default settings defined"""
    element_separator = "*"
    segment_terminator = "~"
    sub_element_separator = ">"
    version = "00501"


class DocumentSettings(object):
    """Current Settings"""
    element_separator = DefaultSettings.element_separator
    segment_terminator = DefaultSettings.segment_terminator
    sub_element_separator = DefaultSettings.sub_element_separator
    version = DefaultSettings.version



from .element import Element
from .envelope import Envelope, GroupEnvelope, InterchangeEnvelope, TransactionSetEnvelope
from .errors import FieldValidationError, IDMismatchError, SegmentCountError
from .interchange import Interchange, InterchangeHeader, InterchangeTrailer
from .segment import Segment



class FieldValidationError(Exception):
    """Exception raised for errors in the input.
    Attributes:
        segment -- segment in which the error occurred
        msg  -- explanation of the error
    """

    def __init__(self, segment, msg):
        self.segment = segment
        self.msg = msg


class IDMismatchError(Exception):
    """Exception raised for errors in the input.
        Attributes:
            segment -- segment in which the error occurred
            msg  -- explanation of the error
        """

    def __init__(self, segment, msg):
        self.segment = segment
        self.msg = msg


class SegmentCountError(Exception):
    """Exception raised for errors in the input.
        Attributes:
            segment -- segment in which the error occurred
            msg  -- explanation of the error
        """

    def __init__(self, segment, msg):
        self.segment = segment
        self.msg = msg


class InvalidFileTypeError(Exception):
    """Exception raised for errors in the input.
    Attributes:
        segment -- segment in which the error occurred
        msg  -- explanation of the error
    """

    def __init__(self, segment, msg):
        self.expr = segment
        self.msg = msg

    def __str__(self):
        return self.msg


class SegmentTerminatorNotFoundError(Exception):
    """Exception raised for errors in the Interchange Header.
    Attributes:
        msg  -- explanation of the error
    """

    def __init__(self, msg):
        self.msg = msg
